type var = string

type binop = Add | Mul | And | Or | Leq | Eq

type exp =
  | Int of int
  | Var of var
  | Binop of binop * exp * exp
  | If of exp * exp * exp

(* Types *)
and ty =
  | TyInt
  | TyBool

type value =
  | VInt of int
  | VBool of bool

and env = (var * value) list

(* Type environment : maps variables to their types *)
type tyenv = (var * ty) list

let rec eval env = function
  | Int k -> VInt k
  | Var v ->
    begin
      try List.assoc v env
      with Not_found ->
        failwith @@ Printf.sprintf "Variable `%s` is not bound" v
    end
  | Binop (op, e1, e2) ->
    begin match op, eval env e1, eval env e2 with
      | Add, VInt v1, VInt v2 -> VInt (v1 + v2)
      | Mul, VInt v1, VInt v2 -> VInt (v1 * v2)
      | And, VBool v1, VBool v2 -> VBool (v1 && v2)
      | Or, VBool v1, VBool v2 -> VBool (v1 || v2)
      | Leq, VInt v1, VInt v2 -> VBool (v1 <= v2)
      | Eq, VInt v1, VInt v2 -> VBool (v1 = v2)
      | _, _, _ -> failwith "Type error!"
    end
  | If (cond, bthen, belse) ->
    begin match eval env cond with
      | VBool true -> eval env bthen
      | VBool false -> eval env belse
      | _ -> failwith "Type error!"
    end

(* Check is the judgement tyenv |- exp : typ is derivable *)
let rec typecheck tyenv exp ty = match exp with
  | Int _ -> ty = TyInt
  | Var v ->
    begin
      try List.assoc v tyenv = ty
      with Not_found ->
        failwith @@ Printf.sprintf "Variable `%s` is not bound" v
    end
  | Binop (Add, e1, e2) | Binop (Mul, e1, e2) ->
    ty = TyInt && typecheck tyenv e1 TyInt && typecheck tyenv e2 TyInt
  | Binop (And, e1, e2) | Binop (Or, e1, e2) ->
    ty = TyBool && typecheck tyenv e1 TyBool && typecheck tyenv e2 TyBool
  | Binop (Leq, e1, e2) | Binop (Eq, e1, e2) ->
    ty = TyBool && typecheck tyenv e1 TyInt && typecheck tyenv e2 TyInt
  | If (cond, bthen, belse) ->
    typecheck tyenv cond TyBool && typecheck tyenv bthen ty && typecheck tyenv belse ty

(* if x <= 0 then -x else x *)
let abs_body = 
  If (Binop (Leq, Var "x", Int 0),
      Binop (Mul, Int (-1), Var "x"),
      Var "x")

(* 0 = (if v then 0 else 1) *)
let negate =
  Binop (Eq,
         Int 0,
         If (Var "v", Int 0, Int 1))

let ill_typed =
  Binop (Eq, (Binop (Leq, Int 0, Int 1)), Int 4)
